
    (*********************************************************************
     *  MODDecoder.pas                                                   *
     *                                                                   *
     *  author    : Milenko Mitrovic                                     *
     *  email     : dcoder@dsp-worx.de                                   *
     *  web       : http://dsp-worx.de                                   *
     *  date      : 31-10-2003                                           *
     *  version   : 0.60                                                 *
     *                                                                   *
     *  The contents of this file are used with permission, subject to   *
     *  the Mozilla Public License Version 1.1 (the "License"); you may  *
     *  not use this file except in compliance with the License. You may *
     *  obtain a copy of the License at                                  *
     *  http://www.mozilla.org/MPL/MPL-1.1.html                          *
     *                                                                   *
     *  Software distributed under the License is distributed on an      *
     *  "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or   *
     *  implied. See the License for the specific language governing     *
     *  rights and limitations under the License.                        *
     *                                                                   *
     *  (C) 2003 Milenko Mitrovic <dcoder@dsp-worx.de>                   *
     *                                                                   *
     *********************************************************************)

unit MODDecoder;

interface

uses
  UnicoWS, Windows, SysUtils, dspConst, dmoConst;

const
  BASS_MUSIC_DECODE       = $200000;
  BASS_MUSIC_RAMP         = $200;
  BASS_MUSIC_CALCLEN      = $8000;
  BASS_UNICODE            = $80000000;
  BASS_SAMPLE_8BITS       = 1;
  BASS_SAMPLE_MONO        = 2;
  BASS_MUSIC_NOSAMPLE     = $100000;
  BASS_MUSIC_POSRESET     = $20000;

type
  TBooleanArray = array[0..1023] of Boolean;
  PBooleanArray = ^TBooleanArray;

  BASS_CHANNELINFO = record
    freq: DWORD;
    chans: DWORD;
    flags: DWORD;
    ctype: DWORD;
  	origres: DWORD;
  end;

  TMODDecoder = class
  private
    BASS_Init : function (device: Integer; freq, flags: DWORD; win: HWND; clsid: PGUID): BOOL; stdcall;
    BASS_MusicLoad : function (mem: BOOL; f: Pointer; offset, length, flags, freq: DWORD): DWORD; stdcall;
    BASS_MusicFree : procedure (handle: DWORD); stdcall;
    BASS_Free : function : BOOL; stdcall;
    BASS_ChannelGetInfo : function (handle: DWORD; var info:BASS_CHANNELINFO):BOOL; stdcall;
    BASS_ChannelGetAttributes : function (handle: DWORD; var freq, volume: DWORD; var pan: Integer): BOOL; stdcall;
    BASS_MusicGetLength : function (handle: DWORD; playlen: BOOL): DWORD; stdcall;
    BASS_ChannelSetPosition : function (handle: DWORD; pos: Int64): BOOL; stdcall;
    BASS_ChannelGetData : function (handle: DWORD; buffer: Pointer; length: DWORD): DWORD; stdcall;
    BASS_MusicGetName : function (handle: DWORD): PChar; stdcall;
    BASS_ChannelSetSync: function (handle: DWORD; stype: DWORD; param: int64; proc: Pointer; user: DWORD): DWORD; stdcall;

    fDuration : Int64;
    fBitsPerSample : integer;
    fBytesPerSample : integer;
    fFrequency : integer;
    fChannels : integer;
    fChan : DWORD;
    fLib : hWnd;

    fIsUMX: Boolean;
    fUMXLen: Int64;                                   
    fUMXaLen: integer;
    fUMXrLen: integer;
    fUMXLast: integer;
    fUMXArr1: PBooleanArray;
    fUMXArr2: PIntegerArray;
    fUMXArr3: PBooleanArray;
    fUMXbr: Boolean;
    fUMXChecking: Boolean;

    procedure SetPosition(Pos : int64);
    procedure GetStreamInfos;
    function CheckUMX(Filename: WideString): Boolean;
    function InitBass(Frequency: integer) : Boolean;
    procedure FreeBass;
    function LoadFile(Params: DWORD; Filename: PWideChar): integer;
  public
    Title : String;
    constructor Create(Filename : PWCHAR; out hr : HRESULT);
    destructor Destroy; override;
    function GetBuffer(Buffer: Pointer; Blocks: Integer) : integer;
  published
    property Duration : Int64 read fDuration;
    property BitsPerSample : integer read fBitsPerSample;
    property BytesPerSample : integer read fBytesPerSample;
    property Frequency : integer read fFrequency;
    property Channels : integer read fChannels;
    property Position : int64 write SetPosition;
  end;

implementation

uses
  MODFilter;

procedure SyncProc(handle: THandle; channel: DWORD; data: DWORD; user: DWORD); stdcall;
var
  tmp: integer;
  i, z: integer;
  dec: TMODDecoder;
begin
  dec := TMODDecoder(user);

  if dec.fUMXChecking then
  begin
    if dec.fUMXrLen < hiword(data) then dec.fUMXrLen := hiword(data);

    tmp := loword(data);
    if tmp <> dec.fUMXLast then
    begin
      dec.fUMXLast := tmp;
      if dec.fUMXArr1[tmp] then
      begin
  //      OutputDebugString(pchar('Repeat ' + inttostr(loword(data))));
        z := -1;
        for i := 0 to dec.fUMXaLen -1 do
        begin
          if not dec.fUMXArr1[i] then
          begin
            z := i;
            break;
          end;
        end;
        if z = -1 then
        begin
  //        OutputDebugString(pchar('End reached ' + inttostr(loword(data))));
          dec.fUMXbr := true;
          exit;
        end else
        begin
  //        OutputDebugString(pchar('Seeking to another position ' + inttostr(z)));
          dec.BASS_ChannelSetPosition(channel,MakeLong(z,0));
          exit;
        end;
      end else
      begin
        dec.fUMXArr1[tmp] := true;
      end;
      for i := 0 to dec.fUMXaLen -1 do
      begin
        if dec.fUMXArr2[i] = -1 then
        begin
          dec.fUMXArr2[i] := tmp;
          break;
        end;
      end;
  //    OutputDebugString(pchar('Position ' + inttostr(loword(data))));
    end;
  end else
  begin
    tmp := loword(data);
    if tmp <> dec.fUMXLast then
    begin
      dec.fUMXLast := tmp;
      if dec.fUMXArr3[tmp] then
      begin
  //      OutputDebugString(pchar('Repeat ' + inttostr(loword(data))));
        z := -1;
        for i := 0 to dec.fUMXaLen -1 do
        begin
          if not dec.fUMXArr3[i] then
          begin
            z := i;
            break;
          end;
        end;
        if z = -1 then
        begin
  //        OutputDebugString(pchar('End reached ' + inttostr(loword(data))));
          dec.fUMXbr := true;
          exit;
        end else
        begin
  //        OutputDebugString(pchar('Seeking to another position ' + inttostr(z)));
          dec.BASS_ChannelSetPosition(channel,MakeLong(z,0));
          exit;
        end;
      end else
      begin
        dec.fUMXArr3[tmp] := true;
      end;
    end;
  end;
end;

constructor TMODDecoder.Create(Filename : PWCHAR; out hr : HRESULT);
var
  i: integer;
begin
  inherited Create;

  Title := '';
  fUMXChecking := False;

  hr := E_FAIL;

  if not CheckUMX(Filename)
    then InitBass(44100);

  fChan := LoadFile(BASS_MUSIC_DECODE or BASS_MUSIC_RAMP or BASS_MUSIC_CALCLEN or BASS_MUSIC_POSRESET, Filename);
  if (fChan = 0) then
  begin
    Exit;
  end else
  begin
    hr := S_OK;
    if fIsUMX then
    begin
      for i := 0 to fUMXaLen -1 do
        BASS_ChannelSetSync(fChan,0,-1,@SyncProc,Cardinal(Self));
    end;
    GetStreamInfos;
  end;
end;

destructor TMODDecoder.Destroy;
begin
  FreeBass;
  if Assigned(fUMXArr1) then FreeMem(fUMXArr1);
  if Assigned(fUMXArr3) then FreeMem(fUMXArr3);
  if Assigned(fUMXArr2) then FreeMem(fUMXArr2);
  inherited Destroy;
end;

procedure TMODDecoder.GetStreamInfos;
var
  lInfo : BASS_CHANNELINFO;
  lFrequency : DWORD;
  lVolume : DWORD;
  lPan : integer;
  lTitle : PChar;
begin
  BASS_ChannelGetInfo(fChan,lInfo);
	if (lInfo.Flags and BASS_SAMPLE_MONO > 0) then fChannels := 1
                                            else fChannels := 2;
  if (lInfo.Flags and BASS_SAMPLE_8BITS > 0) then fBitsPerSample := 8
                                             else fBitsPerSample := 16;
  fBytesPerSample := fBitsPerSample div 8;
	BASS_ChannelGetAttributes(fChan, lFrequency, lVolume, lPan);
  fFrequency := lFrequency;
  if not fIsUMX then
  begin
    fDuration := BASS_MusicGetLength(fChan,True);
    fDuration := fDuration * One_Second div (fFrequency * fBytesPerSample * fChannels);
  end;
  lTitle := BASS_MusicGetName(fChan);
  if Assigned(lTitle) then Title := StrPas(lTitle);
end;

procedure TMODDecoder.SetPosition(Pos : int64);
var
  lPosition : WORD;
  lPosMax : int64;
  t: integer;
  s: integer;
  i: integer;
begin
  lPosMax := fDuration div One_Second;
  lPosition := pos div One_Second;
  if lPosition > lPosMax then lPosition := 0;

  if fIsUMX then
  begin
    t := lPosition * fUMXaLen div fUMXLen;
    s := fUMXArr2[t];
//    OutputDebugString('------------------------------------------------------');
//    OutputDebugString(PChar(inttostr(fUMXrLen)));
//    OutputDebugString(PChar(inttostr(lPosition * (fUMXrLen * fUMXaLen) div fUMXLen mod fUMXrLen)));
    t := lPosition * (fUMXrLen * fUMXaLen) div fUMXLen mod fUMXrLen;
    for i := 0 to fUMXaLen -1 do fUMXArr3[i] := false;
    for i := 0 to s do fUMXArr3[i] := true;
    fUMXLast := s;
    BASS_ChannelSetPosition(fChan,MakeLong(s,t));
  end else
  begin
    BASS_ChannelSetPosition(fChan,MakeLong(lPosition,$FFFF));
  end;
end;

function TMODDecoder.GetBuffer(Buffer: Pointer; Blocks : Integer) : integer;
begin
  Result := BASS_ChannelGetData(fChan, Buffer, Blocks);
end;

function TMODDecoder.InitBass(Frequency: integer) : Boolean;
var
  Buf : array[0..MAX_PATH -1] of Char;
  Path : String;
begin
  Result := False;
  GetModuleFileName(HInstance,Buf,MAX_PATH);
  Path := ExtractFilePath(Buf) + 'bass.dll';
  if FileExists(Path) then fLib := LoadLibrary(PChar(Path))
                      else fLib := LoadLibrary('bass.dll');
  if fLib = 0 then Exit;
  @BASS_Init := GetProcAddress(fLib,'BASS_Init');
  @BASS_MusicLoad := GetProcAddress(fLib,'BASS_MusicLoad');
  @BASS_MusicFree := GetProcAddress(fLib,'BASS_MusicFree');
  @BASS_Free := GetProcAddress(fLib,'BASS_Free');
  @BASS_ChannelGetInfo := GetProcAddress(fLib,'BASS_ChannelGetInfo');
  @BASS_ChannelGetAttributes := GetProcAddress(fLib,'BASS_ChannelGetAttributes');
  @BASS_MusicGetLength := GetProcAddress(fLib,'BASS_MusicGetLength');
  @BASS_ChannelSetPosition := GetProcAddress(fLib,'BASS_ChannelSetPosition');
  @BASS_ChannelGetData := GetProcAddress(fLib,'BASS_ChannelGetData');
  @BASS_MusicGetName := GetProcAddress(fLib,'BASS_MusicGetName');
  @BASS_ChannelSetSync := GetProcAddress(fLib,'BASS_ChannelSetSync');
  BASS_Init(0, Frequency, 0, GetDesktopWindow,nil);
  Result := True;
end;

procedure TMODDecoder.FreeBass;
begin
  if fChan <> 0 then
  begin
    BASS_MusicFree(fChan);
    fChan := 0;
  end;
  if (fLib > 0) and Assigned(Bass_Free) then
  begin
    try
      if (InstanceCount = 0) then
      begin
        BASS_Free;
      end;
    except
      // crashes mplayer2.exe ???
    end;
    FreeLibrary(fLib);
    fLib := 0;
    @BASS_Init := nil;
    @BASS_MusicLoad := nil;
    @BASS_MusicFree := nil;
    @BASS_Free := nil;
    @BASS_ChannelGetInfo := nil;
    @BASS_ChannelGetAttributes := nil;
    @BASS_MusicGetLength := nil;
    @BASS_ChannelSetPosition := nil;
    @BASS_ChannelGetData := nil;
    @BASS_MusicGetName := nil;
    @BASS_ChannelSetSync := nil;
  end;
end;

function TMODDecoder.LoadFile(Params: DWORD; Filename: PWideChar): integer;
var
  P: Pointer;
  hFile: THandle;
  hSize: DWord;
begin
  Result := 0;
  if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) then
  begin
    if not IsMSLULoaded then Exit;

    hFile := UnicoWS.CreateFileW(Filename,GENERIC_READ, FILE_SHARE_READ or FILE_SHARE_WRITE,nil,
                                 OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
    if hFile = INVALID_HANDLE_VALUE then  Exit;

    hSize := GetFileSize(hFile,nil);
    if (hSize = $FFFFFFFF) then
    begin
      CloseHandle(hFile);
      Exit;
    end;

    P := AllocMem(hSize);
    if (P = nil) then
    begin
      CloseHandle(hFile);
      Exit;
    end;

    SetFilePointer(hFile,0,nil,FILE_BEGIN);
    ReadFile(hFile,P^,hSize,hSize,nil);

    CloseHandle(hFile);
    Result := BASS_MusicLoad(True, P, 0, 0, Params,0);
    FreeMem(P);
  end else
  begin
    Result := BASS_MusicLoad(FALSE, FileName, 0, 0, Params or BASS_UNICODE,0);
  end;
end;

function TMODDecoder.CheckUMX(Filename: WideString): Boolean;
var
  i: integer;
  p: pointer;
begin
  Result := False;

  fIsUMX := LowerCase(ExtractFileExt(Filename)) = '.umx';
  if fIsUMX then
  begin
    fUMXChecking := True;
    Result := InitBass(8000);
    if not Result then Exit;
    GetStreamInfos;

    fChan := LoadFile(BASS_MUSIC_DECODE or BASS_MUSIC_NOSAMPLE, PWideChar(Filename));

    if fChan = 0 then
    begin
      Result := False;
      Exit;
    end;


    fUMXLen := BASS_MusicGetLength(fChan,false);
    fUMXaLen := fUMXLen;

    fUMXArr1 := AllocMem(SizeOf(Boolean) * fUMXLen);
    fUMXArr3 := AllocMem(SizeOf(Boolean) * fUMXLen);
    fUMXArr2 := AllocMem(SizeOf(Integer) * fUMXLen);

    p := allocmem(1234);

    for i := 0 to fUMXLen -1 do
    begin
      fUMXArr1[i] := false;
      fUMXArr3[i] := false;
      fUMXArr2[i] := -1;
      BASS_ChannelSetSync(fChan,0,-1,@SyncProc,Cardinal(Self));
    end;

    fUMXrLen := 0;
    fUMXLast := -1;
    fUMXbr := false;
    fUMXLen := 0;

    while true do
    begin
      if fUMXbr then break;
      inc(fUMXLen,BASS_ChannelGetData(fChan,p,1234));
    end;
//    fDuration := fUMXLen * 44100 div 8000;
    fDuration := fUMXLen * One_Second div 8000 div fBytesPerSample div fChannels;
    fUMXLen := fUMXLen div fChannels div (fBitsPerSample div 8) div 8000;
//    fDuration := fUMXLen * One_Second;
    inc(fUMXrLen);

    FreeMem(p);
    FreeBass;
    Result := InitBass(44100);
    fUMXChecking := False;
  end;
end;


end.
